@@ -36,36 +36,7 @@
    in the tree and for manipulating nodes, like replacing, adding and
    deleting nodes.
 
-   $Log: mxml.c,v $
-   Revision 1.10  2005/05/09 18:43:52  ritt
-   Clear error
-
-   Revision 1.9  2005/05/09 09:12:25  ritt
-   Moved strlcpy/strlcat into separate file
-
-   Revision 1.8  2005/04/19 21:43:33  ritt
-   Implemented tree cloning and adding
-
-   Revision 1.7  2005/04/06 11:17:02  ritt
-   Nodes can now have values AND subnodes
-
-   Revision 1.6  2005/03/29 15:11:56  ritt
-   Close element before writing comment
-
-   Revision 1.5  2005/03/29 14:49:50  ritt
-   Fixed compiler warning
-
-   Revision 1.4  2005/03/29 14:48:54  ritt
-   Implemented mxml_write_comment()
-
-   Revision 1.3  2005/03/29 14:37:46  ritt
-   Translate '<' and '&' always in writer
-
-   Revision 1.2  2005/03/29 14:14:38  ritt
-   Implemented mxml_set_translate
-
-   Revision 1.1.1.1  2005/03/29 08:19:45  ritt
-   Imported sources
+   $Id: mxml.c 36 2005-12-20 07:09:34Z ritt $
 
 \********************************************************************/
 
@@ -299,7 +270,6 @@ void mxml_decode(char *str)
          strcpy(p, p+5);
       }
    }
-
    if (str[0] == '\"' && str[strlen(str)-1] == '\"') {
       strcpy(str, str+1);
       str[strlen(str)-1] = 0;
@@ -386,9 +356,8 @@ int mxml_end_element(MXML_WRITER *writer)
       free(writer->stack[writer->level]);
       if (writer->level == 0)
          free(writer->stack);
-      if (write(writer->fh, "/>\n", 3) != 3)
-         return FALSE;
-      return TRUE;
+      strcpy(line, "/>\n");
+      return mxml_write_line(writer, line) == (int)strlen(line);
    }
 
    line[0] = 0;
@@ -758,8 +727,11 @@ int mxml_find_nodes1(PMXML_NODE tree, char *xml_path, PMXML_NODE **nodelist, int
 */
 {
    PMXML_NODE pnode;
-   char *p1, *p2, *p3, node_name[256], condition[256], subnode[256], cond_attr[256], value[256];
-   int i, j, index;
+   char *p1, *p2, *p3, node_name[256], condition[256];
+   char cond_name[MXML_MAX_CONDITION][256], cond_value[MXML_MAX_CONDITION][256];
+   int  cond_type[MXML_MAX_CONDITION];
+   int i, j, k, index, num_cond;
+   int cond_satisfied;
    size_t len;
 
    p1 = xml_path;
@@ -780,8 +752,9 @@ int mxml_find_nodes1(PMXML_NODE tree, char *xml_path, PMXML_NODE **nodelist, int
       memcpy(node_name, p1, len);
       node_name[len] = 0;
       index = 0;
-      subnode[0] = cond_attr[0] = value[0] = 0;
-      if (*p2 == '[') {
+      num_cond = 0;
+      while (*p2 == '[') {
+         cond_name[num_cond][0] = cond_value[num_cond][0] = cond_type[num_cond] = 0;
          p2++;
          if (isdigit(*p2)) {
             /* evaluate [index] */
@@ -802,57 +775,61 @@ int mxml_find_nodes1(PMXML_NODE tree, char *xml_path, PMXML_NODE **nodelist, int
             p2 = strchr(p2, ']')+1;
             if ((p3 = strchr(condition, '=')) != NULL) {
 
-               if (condition[0] == '@') {
-                  strlcpy(cond_attr, condition+1, sizeof(cond_attr));
-                  *strchr(cond_attr, '=') = 0;
-                  while (cond_attr[0] && isspace(cond_attr[strlen(cond_attr)-1]))
-                     cond_attr[strlen(cond_attr)-1] = 0;
-               } else {
-                  strlcpy(subnode, condition, sizeof(subnode));
-                  *strchr(subnode, '=') = 0;
-                  while (subnode[0] && isspace(subnode[strlen(subnode)-1]))
-                     subnode[strlen(subnode)-1] = 0;
-               }
+               if (condition[0] == '@')
+                  cond_type[num_cond] = 1;
+
+               strlcpy(cond_name[num_cond], condition, sizeof(cond_name[num_cond]));
+               *strchr(cond_name[num_cond], '=') = 0;
+               while (cond_name[num_cond][0] && isspace(cond_name[num_cond][strlen(cond_name[num_cond])-1]))
+                  cond_name[num_cond][strlen(cond_name[num_cond])-1] = 0;
+
                p3++;
                while (*p3 && isspace(*p3))
                   p3++;
                if (*p3 == '\"') {
-                  strlcpy(value, p3+1, sizeof(value));
-                  while (value[0] && isspace(value[strlen(value)-1]))
-                     value[strlen(value)-1] = 0;
-                  if (value[0] && value[strlen(value)-1] == '\"')
-                     value[strlen(value)-1] = 0;
+                  strlcpy(cond_value[num_cond], p3+1, sizeof(cond_value[num_cond]));
+                  while (cond_value[num_cond][0] && isspace(cond_value[num_cond][strlen(cond_value[num_cond])-1]))
+                     cond_value[num_cond][strlen(cond_value[num_cond])-1] = 0;
+                  if (cond_value[num_cond][0] && cond_value[num_cond][strlen(cond_value[num_cond])-1] == '\"')
+                     cond_value[num_cond][strlen(cond_value[num_cond])-1] = 0;
                } else if (*p3 == '\'') {
-                  strlcpy(value, p3+1, sizeof(value));
-                  while (value[0] && isspace(value[strlen(value)-1]))
-                     value[strlen(value)-1] = 0;
-                  if (value[0] && value[strlen(value)-1] == '\'')
-                     value[strlen(value)-1] = 0;
+                  strlcpy(cond_value[num_cond], p3+1, sizeof(cond_value[num_cond]));
+                  while (cond_value[num_cond][0] && isspace(cond_value[num_cond][strlen(cond_value[num_cond])-1]))
+                     cond_value[num_cond][strlen(cond_value[num_cond])-1] = 0;
+                  if (cond_value[num_cond][0] && cond_value[num_cond][strlen(cond_value[num_cond])-1] == '\'')
+                     cond_value[num_cond][strlen(cond_value[num_cond])-1] = 0;
                } else {
-                  strlcpy(value, p3, sizeof(value));
-                  while (value[0] && isspace(value[strlen(value)-1]))
-                     value[strlen(value)-1] = 0;
+                  strlcpy(cond_value[num_cond], p3, sizeof(cond_value[num_cond]));
+                  while (cond_value[num_cond][0] && isspace(cond_value[num_cond][strlen(cond_value[num_cond])-1]))
+                     cond_value[num_cond][strlen(cond_value[num_cond])-1] = 0;
                }
+               num_cond++;
             }
          }
       }
 
       for (i=j=0 ; i<pnode->n_children ; i++) {
-         if (subnode[0]) {
-            /* search subnode */
-            for (j=0 ; j<pnode->child[i].n_children ; j++)
-               if (strcmp(pnode->child[i].child[j].name, subnode) == 0)
-                  if (strcmp(pnode->child[i].child[j].value, value) == 0)
-                     if (!mxml_add_resultnode(pnode->child+i, p2, nodelist, found))
-                        return 0;
-
-         } else if (cond_attr[0]) {
-            /* search node with attribute */
-            if (strcmp(pnode->child[i].name, node_name) == 0)
-               if (mxml_get_attribute(pnode->child+i, cond_attr) &&
-                  strcmp(mxml_get_attribute(pnode->child+i, cond_attr), value) == 0)
-                  if (!mxml_add_resultnode(pnode->child+i, p2, nodelist, found))
-                     return 0;
+         if (num_cond) {
+            cond_satisfied = 0;
+            for (k=0;k<num_cond;k++) {
+               if (cond_type[k]) {
+                  /* search node with attribute */
+                  if (strcmp(pnode->child[i].name, node_name) == 0)
+                     if (mxml_get_attribute(pnode->child+i, cond_name[k]) &&
+                        strcmp(mxml_get_attribute(pnode->child+i, cond_name[k]), cond_value[k]) == 0)
+                        cond_satisfied++;
+               }
+               else {
+                  /* search subnode */
+                  for (j=0 ; j<pnode->child[i].n_children ; j++)
+                     if (strcmp(pnode->child[i].child[j].name, cond_name[k]) == 0)
+                        if (strcmp(pnode->child[i].child[j].value, cond_value[k]) == 0)
+                           cond_satisfied++;
+               }
+            }
+            if (cond_satisfied==num_cond)
+               if (!mxml_add_resultnode(pnode->child+i, p2, nodelist, found))
+                  return 0;
          } else {
             if (strcmp(pnode->child[i].name, node_name) == 0)
                if (index == 0 || ++j == index)
@@ -1127,7 +1104,7 @@ PMXML_NODE mxml_parse_buffer(char *buf, char *error, int error_size)
 {
    char node_name[256], attrib_name[256], attrib_value[1000];
    char *p, *pv;
-   int i, line_number;
+   int i,j, line_number;
    PMXML_NODE root, ptree, pnew;
    int end_element;
    size_t len;
@@ -1146,7 +1123,7 @@ PMXML_NODE mxml_parse_buffer(char *buf, char *error, int error_size)
          end_element = FALSE;
 
          /* found new element */
-         *p++;
+         p++;
          while (*p && isspace(*p)) {
             if (*p == '\n')
                line_number++;
@@ -1206,6 +1183,28 @@ PMXML_NODE mxml_parse_buffer(char *buf, char *error, int error_size)
 
             p += 2;
 
+         } else if (strncmp(p, "!DOCTYPE", 8) == 0 ) {
+
+            /* found !DOCTYPE element , skip it */
+            p += 8;
+            if (strstr(p, ">") == NULL)
+               return read_error(HERE, "Unterminated !DOCTYPE element");
+
+            j = 0;
+            while (*p && (*p != '>' || j > 0)) {
+               if (*p == '\n')
+                  line_number++;
+               else if (*p == '<')
+                  j++;
+               else if (*p == '>')
+                  j--;
+               p++;
+            }
+            if (!*p)
+               return read_error(HERE, "Unexpected end of file");
+
+            p++;
+
          } else {
             
             /* found normal element */
@@ -1404,6 +1403,449 @@ PMXML_NODE mxml_parse_buffer(char *buf, char *error, int error_size)
 
 /*------------------------------------------------------------------*/
 
+int mxml_parse_entity(char **buf, char *file_name, char *error, int error_size)
+/* parse !ENTYTY entries of XML files and replace with references. Return 0
+   in case of no errors, return error description. Optional file_name is used
+   for error reporting if called from mxml_parse_file() */
+{
+   char *p;
+   char *pv;
+   char delimiter;
+   int i, j, k, line_number;
+   char *replacement;
+   char entity_name[MXML_MAX_ENTITY][256];
+   char entity_reference_name[MXML_MAX_ENTITY][256];
+   char *entity_value[MXML_MAX_ENTITY];
+   int entity_type[MXML_MAX_ENTITY];    /* internal or external */
+   int nentity;
+   int fh, length, len;
+   char *buffer;
+   PMXML_NODE root = mxml_create_root_node();   /* dummy for 'HERE' */
+   int ip;                      /* counter for entity value */
+   char directoryname[FILENAME_MAX];
+   char filename[FILENAME_MAX];
+
+   for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+      entity_value[ip] = NULL;
+
+   line_number = 1;
+   nentity = -1;
+
+   if (!buf || !(*buf) || !strlen(*buf))
+      return 0;
+
+   strcpy(directoryname, file_name);
+   mxml_dirname(directoryname);
+
+   /* copy string to temporary space */
+   buffer = (char *) malloc(strlen(*buf) + 1);
+   if (buffer == NULL) {
+      read_error(HERE, "Cannot allocate memory.");
+      free(buffer);
+      for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+         free(entity_value[ip]);
+      return 1;
+   }
+   strcpy(buffer, *buf);
+
+   p = strstr(buffer, "!DOCTYPE");
+   if (p == NULL) {             /* no entities */
+      mxml_free_tree(root);
+      free(buffer);
+      for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+         free(entity_value[ip]);
+      return 0;
+   }
+
+   pv = strstr(p, "[");
+   if (pv == NULL) {            /* no entities */
+      mxml_free_tree(root);
+      free(buffer);
+      for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+         free(entity_value[ip]);
+      return 0;
+   }
+
+   p = pv + 1;
+
+   /* search !ENTITY */
+   do {
+      if (*p == ']')
+         break;
+
+      if (*p == '<') {
+
+         /* found new entity */
+         p++;
+         while (*p && isspace(*p)) {
+            if (*p == '\n')
+               line_number++;
+            p++;
+         }
+         if (!*p) {
+            read_error(HERE, "Unexpected end of file");
+            free(buffer);
+            for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+               free(entity_value[ip]);
+            return 1;
+         }
+
+         if (strncmp(p, "!--", 3) == 0) {
+            /* found comment */
+            p += 3;
+            if (strstr(p, "-->") == NULL) {
+               read_error(HERE, "Unterminated comment");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            while (strncmp(p, "-->", 3) != 0) {
+               if (*p == '\n')
+                  line_number++;
+               p++;
+            }
+            p += 3;
+         }
+
+         else if (strncmp(p, "!ENTITY", 7) == 0) {
+            /* found entity */
+            nentity++;
+            if (nentity >= MXML_MAX_ENTITY) {
+               read_error(HERE, "Too much entities");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            pv = p + 7;
+            while (*pv == ' ')
+               pv++;
+
+            /* extract entity name */
+            p = pv;
+
+            while (*p && isspace(*p) && *p != '<' && *p != '>') {
+               if (*p == '\n')
+                  line_number++;
+               p++;
+            }
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*p == '<' || *p == '>') {
+               read_error(HERE, "Unexpected \'%c\' inside !ENTITY", *p);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            pv = p;
+            while (*pv && !isspace(*pv) && *pv != '<' && *pv != '>')
+               pv++;
+
+            if (!*pv) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*pv == '<' || *pv == '>') {
+               read_error(HERE, "Unexpected \'%c\' inside entity \"%s\"", *pv, &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            len = (size_t) pv - (size_t) p;
+
+            entity_name[nentity][0] = '&';
+            i = 1;
+            entity_name[nentity][i] = 0;
+            while (*p && !isspace(*p) && *p != '/' && *p != '>' && *p != '<' && i < 253)
+               entity_name[nentity][i++] = *p++;
+            entity_name[nentity][i++] = ';';
+            entity_name[nentity][i] = 0;
+
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*p == '<') {
+               read_error(HERE, "Unexpected \'<\' inside entity \"%s\"", &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            /* extract replacement or SYSTEM */
+            while (*p && isspace(*p)) {
+               if (*p == '\n')
+                  line_number++;
+               p++;
+            }
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*p == '>') {
+               read_error(HERE, "Unexpected \'>\' inside entity \"%s\"", &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            /* check if SYSTEM */
+            if (strncmp(p, "SYSTEM", 6) == 0) {
+               entity_type[nentity] = EXTERNAL_ENTITY;
+               p += 6;
+            } else {
+               entity_type[nentity] = INTERNAL_ENTITY;
+            }
+
+            /* extract replacement */
+            while (*p && isspace(*p)) {
+               if (*p == '\n')
+                  line_number++;
+               p++;
+            }
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*p == '>') {
+               read_error(HERE, "Unexpected \'>\' inside entity \"%s\"", &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            if (*p != '\"' && *p != '\'') {
+               read_error(HERE, "Replacement was not found for entity \"%s\"", &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            delimiter = *p;
+            p++;
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            pv = p;
+            while (*pv && *pv != delimiter)
+               pv++;
+
+            if (!*pv) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            if (*pv == '<') {
+               read_error(HERE, "Unexpected \'%c\' inside entity \"%s\"", *pv, &entity_name[nentity][1]);
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            len = (size_t) pv - (size_t) p;
+            replacement = (char *) malloc(len + 1);
+            if (replacement == NULL) {
+               read_error(HERE, "Cannot allocate memory.");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+
+            memcpy(replacement, p, len);
+            replacement[len] = 0;
+            mxml_decode(replacement);
+
+            if (entity_type[nentity] == EXTERNAL_ENTITY) {
+               strcpy(entity_reference_name[nentity], replacement);
+            } else {
+               entity_value[nentity] = (char *) malloc(strlen(replacement));
+               if (entity_value[nentity] == NULL) {
+                  read_error(HERE, "Cannot allocate memory.");
+                  free(buffer);
+                  for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                     free(entity_value[ip]);
+                  return 1;
+               }
+               strcpy(entity_value[nentity], replacement);
+            }
+            free(replacement);
+
+            p = pv;
+            while (*p && isspace(*p)) {
+               if (*p == '\n')
+                  line_number++;
+               p++;
+            }
+            if (!*p) {
+               read_error(HERE, "Unexpected end of file");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+         }
+      }
+
+      /* go to next element */
+      while (*p && *p != '<') {
+         if (*p == '\n')
+            line_number++;
+         p++;
+      }
+   } while (*p);
+   nentity++;
+
+   /* read external file */
+   for (i = 0; i < nentity; i++) {
+      if (entity_type[i] == EXTERNAL_ENTITY) {
+         if ( entity_reference_name[i][0] == DIR_SEPARATOR ) /* absolute path */
+            strcpy(filename, entity_reference_name[i]);
+         else /* relative path */
+            sprintf(filename, "%s%c%s", directoryname, DIR_SEPARATOR, entity_reference_name[i]);
+         fh = open(filename, O_RDONLY | O_TEXT, 0644);
+
+         if (fh == -1) {
+            entity_value[i] =
+                (char *) malloc(strlen(entity_reference_name[i]) + strlen("<!--  is missing -->") + 1);
+            if (entity_value[i] == NULL) {
+               read_error(HERE, "Cannot allocate memory.");
+               free(buffer);
+               for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                  free(entity_value[ip]);
+               return 1;
+            }
+            sprintf(entity_value[i], "<!-- %s is missing -->", entity_reference_name[i]);
+         } else {
+            length = lseek(fh, 0, SEEK_END);
+            lseek(fh, 0, SEEK_SET);
+            if (length == 0) {
+               entity_value[i] = (char *) malloc(1);
+               if (entity_value[i] == NULL) {
+                  read_error(HERE, "Cannot allocate memory.");
+                  free(buffer);
+                  for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                     free(entity_value[ip]);
+                  return 1;
+               }
+               entity_value[i][0] = 0;
+            } else {
+               entity_value[i] = (char *) malloc(length);
+               if (entity_value[i] == NULL) {
+                  read_error(HERE, "Cannot allocate memory.");
+                  free(buffer);
+                  for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                     free(entity_value[ip]);
+                  return 1;
+               }
+
+               /* read complete file at once */
+               length = read(fh, entity_value[i], length);
+               entity_value[i][length - 1] = 0;
+               close(fh);
+
+               /* recursive parse */
+               if (mxml_parse_entity(&entity_value[i], filename, error, error_size) != 0) {
+                  mxml_free_tree(root);
+                  free(buffer);
+                  for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+                     free(entity_value[ip]);
+                  return 1;
+               }
+            }
+         }
+      }
+   }
+
+   /* count length of output string */
+   length = strlen(buffer);
+   for (i = 0; i < nentity; i++) {
+      p = buffer;
+      while (1) {
+         pv = strstr(p, entity_name[i]);
+         if (pv) {
+            length += strlen(entity_value[i]) - strlen(entity_name[i]);
+            p = pv + 1;
+         } else {
+            break;
+         }
+      }
+   }
+
+   /* re-allocate memory */
+   free(*buf);
+   *buf = (char *) malloc(length + 1);
+   if (*buf == NULL) {
+      read_error(HERE, "Cannot allocate memory.");
+      free(buffer);
+      for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+         free(entity_value[ip]);
+      return 1;
+   }
+
+   /* replace entities */
+   p = buffer;
+   pv = *buf;
+   do {
+      if (*p == '&') {
+         /* found entity */
+         for (j = 0; j < nentity; j++) {
+            if (strncmp(p, entity_name[j], strlen(entity_name[j])) == 0) {
+               for (k = 0; k < (int) strlen(entity_value[j]); k++)
+                  *pv++ = entity_value[j][k];
+               p += strlen(entity_name[j]);
+               break;
+            }
+         }
+      }
+      *pv++ = *p++;
+   } while (*p);
+   *pv = 0;
+
+   free(buffer);
+   for (ip = 0; ip < MXML_MAX_ENTITY; ip++)
+      free(entity_value[ip]);
+
+   mxml_free_tree(root);
+   return 0;
+}
+
+/*------------------------------------------------------------------*/
+
 PMXML_NODE mxml_parse_file(char *file_name, char *error, int error_size)
 /* parse a XML file and convert it into a tree of MXML_NODE's. Return NULL
    in case of an error, return error description */
@@ -1440,6 +1882,11 @@ PMXML_NODE mxml_parse_file(char *file_name, char *error, int error_size)
    buf[length] = 0;
    close(fh);
 
+   if (mxml_parse_entity(&buf, file_name, error, error_size) != 0) {
+      free(buf);
+      return NULL;
+   }
+
    root = mxml_parse_buffer(buf, error, error_size);
 
    free(buf);
@@ -1619,3 +2066,78 @@ void mxml_test()
    mxml_free_tree(tree);
 }
 */
+
+/*------------------------------------------------------------------*/
+/* mxml_basename deletes any prefix ending with the last slash '/' character
+   present in path. mxml_dirname deletes the filename portion, beginning with
+   the last slash '/' character to the end of path. Followings are examples
+   from these functions
+
+    path               dirname   basename
+    "/"                "/"       ""
+    "."                "."       "."
+    ""                 ""        ""
+    "/test.txt"        "/"       "test.txt"
+    "path/to/test.txt" "path/to" "test.txt"
+    "test.txt          "."       "test.txt"
+
+   Under Windows, '\\' and ':' are recognized ad separator too. */
+
+void mxml_basename(char *path)
+{
+   char str[FILENAME_MAX];
+   char *p;
+   char *name;
+
+   if (path) {
+      strcpy(str, path);
+      p = str;
+      name = str;
+      while (1) {
+         if (*p == 0)
+            break;
+         if (*p == '/'
+#ifdef _MSC_VER
+             || *p == ':' || *p == '\\'
+#endif
+             )
+            name = p + 1;
+         p++;
+      }
+      strcpy(path, name);
+   }
+
+   return;
+}
+
+void mxml_dirname(char *path)
+{
+   char *p;
+#ifdef _MSC_VER
+   char *pv;
+#endif
+
+   if (!path || strlen(path) == 0)
+      return;
+
+   p = strrchr(path, '/');
+#ifdef _MSC_VER
+   pv = strrchr(path, ':');
+   if (pv > p)
+      p = pv;
+   pv = strrchr(path, '\\');
+   if (pv > p)
+      p = pv;
+#endif
+
+   if (p == 0)                  /* current directory */
+      strcpy(path, ".");
+   else if (p == path)          /* root directory */
+      sprintf(path, "%c", *p);
+   else
+      *p = 0;
+
+   return;
+}
+
+/*------------------------------------------------------------------*/
